//  SuperTux
//  Copyright (C) 2023 mrkubax10 <mrkubax10@onet.pl>
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#pragma once

#include "gui/menu.hpp"

class AddonMenu;

class AddonFileInstallMenu final : public Menu
{
public:
  AddonFileInstallMenu(AddonMenu* addon_menu);

  virtual void refresh() override;
  virtual void menu_action(MenuItem& item) override {}

  virtual void event(const SDL_Event& event) override;

private:
  AddonMenu* m_addon_menu;

private:
  AddonFileInstallMenu(const AddonFileInstallMenu&) = delete;
  AddonFileInstallMenu& operator=(const AddonFileInstallMenu&) = delete;
};
